<?php

/**
 * Contao Open Source CMS
 * 
 * Copyright (C) 2005-2013 Leo Feyer
 * 
 * @copyright	Tim Gatzky 2013, Premium Contao Webworks, Premium Contao Themes
 * @author		Tim Gatzky <info@tim-gatzky.de>
 * @package		pct_customelements
 * @link		http://contao.org
 */

/**
 * Namespace
 */
namespace PCT\CustomElements\Core;

/**
 * Imports
 */
use PCT\CustomElements\Core\CustomElements as CustomElements;
use PCT\CustomElements\Models\PluginModel as PluginModel;

/**
 * Class file
 * PluginFactory
 * Manage Plugins for CustomElements
 */
class PluginFactory
{
	/**
	 * Object instance (Singleton)
	 * @var object
	 */
	protected static $objInstance;

	/**
	 * Return the current object instance (Singleton)
	 * @return object
	 */
	public static function getInstance()
	{
		if (static::$objInstance === null)
		{
			static::$objInstance = new static();
		}
		return static::$objInstance;
	}
	
	
	/**
	 * Return all registed plugins as array
	 * @return array
	 */
	public static function getRegistered()
	{
		if(empty($GLOBALS['PCT_CUSTOMELEMENTS']['PLUGINS']) || !isset($GLOBALS['PCT_CUSTOMELEMENTS']['PLUGINS']))
		{
			return array();
		}
		
		$arrReturn = array();
		foreach($GLOBALS['PCT_CUSTOMELEMENTS']['PLUGINS'] as $name => $arrDef)
		{
			$arrReturn[] = $name;
		}
		
		return $arrReturn;
	}
	
	
	/**
	 * Return all available plugins as array after comparing to the the plugins already registered
	 * @return array
	 */
	public static function getAvailable()
	{
		$arrRegistered = static::getRegistered();
		$arrPublished = static::findAllPublished();
		
		$arrReturn = array_unique(array_diff($arrRegistered, $arrPublished));
		
		return $arrReturn;
	}
	
	
	/**
	 * Get all published plugins as array
	 * @return array
	 */
	public static function findAllPublished()
	{
		if( !\Contao\Database::getInstance()->tableExists(PluginModel::getTable()) )
		{
			return array();
		}
		
		$objCollection = PluginModel::findAllPublished();
		if($objCollection === null)
		{
			return array();
		}
		
		return $objCollection->fetchEach('plugin');
	}
	
	
	/**
	 * Get all published plugins with their definition as array
	 * @return array 
	 */
	public static function findAllPublishedWithDefinition()
	{
		$arrPlugins = array();
		if(is_array(Cache::getPlugins()))
		{
			$arrPlugins = Cache::getPlugins();
		}
		else
		{
			$arrPlugins = static::findAllPublished();
		}
		
		if(empty($arrPlugins))
		{
			return array();
		}
			
		$arrReturn = array();
		foreach($arrPlugins as $name)
		{
			if(!static::hasAccess( PluginModel::findByAlias($name) ))
			{
				continue;
			}
			
			$arrReturn[$name] = $GLOBALS['PCT_CUSTOMELEMENTS']['PLUGINS'][$name];
		}
		
		return $arrReturn;
	}
	
	
	/**
	 * Check if the current backend user has access to the plugin
	 * @param object	\PCT\CustomElements\Models\PluginModel
	 * @return boolean
	 */
	public static function hasAccess($objModel)
	{
		$blnInstallTool = false;
		if(strlen(strpos(\Contao\Environment::getInstance()->scriptName, '/contao/install.php')) > 0 || strlen(strpos(\Contao\Environment::getInstance()->requestUri, '/contao/install')) > 0)
		{
			$blnInstallTool = true;
		}
		
		if(TL_MODE == 'FE' || (boolean)$objModel->proctected === false  || $blnInstallTool)
		{
			return true;
		}
		
		$objUser = \Contao\BackendUser::getInstance();
		if(!$objUser->id)
		{
			$objUser->authenticate();
		}
		
		if($objUser->isAdmin)
		{
			return true;
		}
		
		// check if user has group access
		$usergroups = \deserialize($objModel->user_groups);
		if(!empty($usergroups) && is_array($usergroups))
		{
			foreach($usergroups as $group)
			{
				if(in_array($group, $objUser->groups))
				{
					return true;
				}
			}
		}
		
		// check if user itself has access
		$users = \deserialize($objModel->users);
		if(!empty($users) && is_array($users))
		{
			if(in_array($users, $objUser->id))
			{
				return true;
			}
		}
		
		return false;
	}
	
		
	/**
	 * Get all available plugins check for permission
	 * @return array
	 */
	public static function getActivePlugins()
	{
		if(is_array(Cache::getPlugins()))
		{
			return Cache::getPlugins();
		}
		
		$arrPlugins = static::findAllPublished();
		if(empty($arrPlugins))
		{
			Cache::addPlugins(array());
			return array();
		}
		
		$arrReturn = array();
		
		if(TL_MODE == 'FE')
		{
			$arrReturn = $arrPlugins;	
		}
		// check if backend user has rights to access the plugin
		else if(TL_MODE == 'BE')
		{
			$arrReturn = array();
			foreach($arrPlugins as $name)
			{
				if( !static::hasAccess( PluginModel::findByAlias($name)) )
				{
					continue;
				}
				$arrReturn[] = $name;
			}
		}
		
		// add to cache
		Cache::addPlugins($arrReturn);
		
		return $arrReturn;
	}
	
	
	/**
	 * Return all excluded custom elements listed by plugin
	 * @param string	Name of the plugin
	 * @return array
	 */
	public static function getExcludes($strPlugin)
	{
		$arrCache = \PCT\CustomElements\Core\Cache::getData();
		if(isset($arrCache['PLUGINS_EXCLUDES'][$strPlugin]))
		{
			return $arrCache['PLUGINS_EXCLUDES'][$strPlugin];
		}
		
		$objModel = PluginModel::findByAlias($strPlugin);
			
		$arrReturn = array();
		if($objModel->customelements) 
		{
			$arrReturn = \deserialize($objModel->customelements);
		}
		
		// add to cache
		$arrCache['PLUGINS_EXCLUDES'][$strPlugin] = $arrReturn;
		\PCT\CustomElements\Core\Cache::setData($arrCache);
		
		return $arrReturn;	
	}
	
}